"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const R = __importStar(require("ramda"));
const Rx = __importStar(require("rxjs"));
const cli_table_1 = __importDefault(require("cli-table"));
const commander_1 = __importDefault(require("commander"));
const types_1 = require("./types");
const radar_relay_1 = require("./exchanges/radar-relay");
const kraken_1 = require("./exchanges/kraken");
const kyber_1 = require("./exchanges/kyber");
const ethfinex_1 = require("./exchanges/ethfinex");
const token_1 = require("@melonproject/token-math/token");
const price_1 = require("@melonproject/token-math/price");
const operators_1 = require("rxjs/operators");
const aggregate_1 = require("./exchanges/aggregate");
const debug = require('debug')('exchange-aggregator');
const displayPrice = (price, decimals) => {
    const value = price_1.toFixed(price, decimals);
    const base = price.base.token.symbol;
    const quote = price.quote.token.symbol;
    return `${value} ${base}/${quote}`;
};
exports.exchangeOrderObservableCreators = {
    [types_1.Exchange.RADAR_RELAY]: (options) => radar_relay_1.observeRadarRelay(options),
    [types_1.Exchange.KRAKEN]: (options) => kraken_1.observeKraken(options),
    [types_1.Exchange.KYBER]: (options) => kyber_1.observeKyber(options),
    [types_1.Exchange.ETHFINEX]: (options) => ethfinex_1.observeEthfinex(options),
};
exports.createExchangeOrderObservable = R.curry((options, exchange) => {
    return exports.exchangeOrderObservableCreators[exchange](options);
});
exports.createExchangeOrderObservables = R.curry((options, exchanges = [types_1.Exchange.RADAR_RELAY]) => {
    return exchanges.map(exports.createExchangeOrderObservable(options));
});
commander_1.default.description('The exchange-aggregator testing tool.');
commander_1.default
    .command('orders [exchanges...]')
    .option('-n, --network <network>', 'The network (KOVAN/MAINNET).', /(MAINNET|KOVAN)/, 'MAINNET')
    .option('-b, --base <symbol>', 'The base token symbol.', /[A-Z]{3,4}/, 'ZRX')
    .option('-q, --quote <symbol>', 'The quote token symbol.', /[A-Z]{3,4}/, 'ETH')
    .description('Retrieve orders from the given exchanges')
    .action((args, options) => {
    const supported = Object.keys(types_1.Exchange);
    const invalid = R.difference(args, supported);
    if (invalid && invalid.length) {
        console.error(`Invalid exchanges %s. Supported exchanges are %s.`, invalid.join(', '), supported.join(', '));
        process.exit(1);
    }
    const exchanges = args.length ? args : supported;
    const opts = {
        network: types_1.Network[options.network],
        pair: {
            base: token_1.createToken(options.base),
            quote: token_1.createToken(options.quote),
        },
    };
    debug('Aggregating orderbook for %s.', exchanges.join(', '));
    const observables = exports.createExchangeOrderObservables(opts, exchanges);
    Rx.merge(...observables)
        .pipe(operators_1.scan(aggregate_1.aggregateOrderbook, aggregate_1.initializeOrderbook(opts)), operators_1.tap(orderbook => {
        const bids = orderbook.bids.length;
        const asks = orderbook.asks.length;
        debug('Aggregated orderbook with %s bids and %s asks.', bids, asks);
    }), operators_1.throttleTime(5000))
        .subscribe(orderbook => {
        const style = {
            head: ['ID', 'Exchange', 'Price'],
            colWidths: [25, 15, 25],
            chars: {
                top: '═',
                'top-mid': '╤',
                'top-left': '╔',
                'top-right': '╗',
                bottom: '═',
                'bottom-mid': '╧',
                'bottom-left': '╚',
                'bottom-right': '╝',
                left: '║',
                'left-mid': '╟',
                mid: '─',
                'mid-mid': '┼',
                right: '║',
                'right-mid': '╢',
                middle: '│',
            },
        };
        const bids = new cli_table_1.default(style);
        orderbook.bids.forEach(value => {
            const price = displayPrice(value.trade);
            bids.push([value.id || '???', value.exchange, price]);
        });
        const asks = new cli_table_1.default(style);
        orderbook.asks.forEach(value => {
            const price = displayPrice(value.trade);
            asks.push([value.id || '???', value.exchange, price]);
        });
        console.log('Asks');
        console.log(asks.toString());
        console.log('Bids');
        console.log(bids.toString());
    }, error => {
        debug('Orderbook aggregation failed.', error);
    });
    process.stdin.resume();
});
commander_1.default.parse(process.argv);
