"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const R = __importStar(require("ramda"));
const Rx = __importStar(require("rxjs"));
const axios_1 = __importDefault(require("axios"));
const operators_1 = require("rxjs/operators");
const webSocket_1 = require("rxjs/webSocket");
const isomorphic_ws_1 = __importDefault(require("isomorphic-ws"));
const types_1 = require("@radarrelay/types");
const types_2 = require("../../types");
const price_1 = require("@melonproject/token-math/price");
const quantity_1 = require("@melonproject/token-math/quantity");
const debug_1 = require("../debug");
const debug = require('debug')('exchange-aggregator:radar-relay');
const subscribeMessage = (options, id) => {
    const baseToken = cleanToken(options.pair.base.symbol);
    const quoteToken = cleanToken(options.pair.quote.symbol);
    const message = {
        type: types_1.WebsocketRequestType.SUBSCRIBE,
        topic: types_1.WebsocketRequestTopic.BOOK,
        market: `${baseToken}-${quoteToken}`,
        requestId: id,
    };
    return message;
};
const unSubscribeMessage = (options) => {
    const baseToken = cleanToken(options.pair.base.symbol);
    const quoteToken = cleanToken(options.pair.quote.symbol);
    const message = {
        type: types_1.WebsocketRequestType.UNSUBSCRIBE,
        topic: types_1.WebsocketRequestTopic.BOOK,
        market: `${baseToken}-${quoteToken}`,
    };
    return message;
};
const normalizeOrder = R.curryN(2, (options, order) => {
    const base = parseFloat(order.remainingBaseTokenAmount);
    const quote = parseFloat(order.remainingQuoteTokenAmount);
    const price = price_1.createPrice(quantity_1.createQuantity(options.pair.base, base), quantity_1.createQuantity(options.pair.quote, quote));
    return {
        id: order.orderHash,
        type: order.type,
        exchange: types_2.Exchange.RADAR_RELAY,
        trade: price,
    };
});
const normalizeNewOrderEvent = R.curryN(2, (options, event) => ({
    event: types_2.NormalizedMessageType.ADD,
    exchange: types_2.Exchange.RADAR_RELAY,
    id: event.order.orderHash,
    order: normalizeOrder(options, event.order),
}));
const normalizeFillOrderEvent = R.curryN(2, (options, event) => ({
    event: types_2.NormalizedMessageType.ADD,
    exchange: types_2.Exchange.RADAR_RELAY,
    id: event.order.orderHash,
    order: normalizeOrder(options, event.order),
}));
const normalizeCancelOrderEvent = (event) => ({
    event: types_2.NormalizedMessageType.REMOVE,
    exchange: types_2.Exchange.RADAR_RELAY,
    id: event.orderHash,
});
const normalizeRemoveOrderEvent = (event) => ({
    event: types_2.NormalizedMessageType.REMOVE,
    exchange: types_2.Exchange.RADAR_RELAY,
    id: event.orderHash,
});
const normalizeSnapshotEvent = R.curryN(2, (options, book) => ({
    event: types_2.NormalizedMessageType.SNAPSHOT,
    exchange: types_2.Exchange.RADAR_RELAY,
    orders: [].concat(book.asks.map(normalizeOrder(options)), book.bids.map(normalizeOrder(options))),
}));
exports.isSnapshotEvent = R.allPass([R.has('asks'), R.has('bids')]);
exports.isOrderEvent = R.allPass([R.has('event'), R.has('action')]);
exports.isNewOrderEvent = R.allPass([
    exports.isOrderEvent,
    R.propEq('action', 'NEW'),
]);
exports.isFillOrderEvent = R.allPass([
    exports.isOrderEvent,
    R.propEq('action', 'FILL'),
]);
exports.isCancelOrderEvent = R.allPass([
    exports.isOrderEvent,
    R.propEq('action', 'CANCEL'),
]);
exports.isRemoveOrderEvent = R.allPass([
    exports.isOrderEvent,
    R.propEq('action', 'REMOVE'),
]);
const websocketConnections = {
    [types_2.Network.KOVAN]: undefined,
    [types_2.Network.MAINNET]: undefined,
};
const getWebsocketConnection = (options) => {
    if (typeof websocketConnections[options.network] !== 'undefined') {
        return websocketConnections[options.network];
    }
    const open$ = new Rx.Subject();
    const close$ = new Rx.Subject();
    const url = getWebsocketUrl(options);
    const connection$ = webSocket_1.webSocket({
        WebSocketCtor: isomorphic_ws_1.default,
        closeObserver: close$,
        openObserver: open$,
        url,
    });
    open$.subscribe(() => {
        debug('Opening connection.');
    });
    close$.subscribe(() => {
        debug('Closing connection.');
    });
    return (websocketConnections[options.network] = connection$);
};
let requestId = 0;
const getRequestId = () => {
    return (requestId = requestId + 1);
};
const getWebsocketUrl = (options) => {
    switch (options.network) {
        case types_2.Network.KOVAN:
            return 'wss://ws.kovan.radarrelay.com/v2';
        case types_2.Network.MAINNET:
            return 'wss://ws.radarrelay.com/v2';
        default:
            throw new Error('Invalid network.');
    }
};
const cleanToken = (token) => {
    switch (token) {
        case 'ETH':
            return 'WETH';
        default:
            return token;
    }
};
const getHttpUrl = (options) => {
    const base = cleanToken(options.pair.base.symbol);
    const quote = cleanToken(options.pair.quote.symbol);
    switch (options.network) {
        case types_2.Network.KOVAN: {
            const prefix = 'https://api.kovan.radarrelay.com';
            return `${prefix}/v2/markets/${base}-${quote}/book`;
        }
        case types_2.Network.MAINNET: {
            const prefix = 'https://api.kovan.radarrelay.com';
            return `${prefix}/v2/markets/${base}-${quote}/book`;
        }
        default:
            throw new Error('Invalid network.');
    }
};
exports.observeRadarRelay = (options) => {
    try {
        const id = getRequestId();
        const ws$ = getWebsocketConnection(options);
        const socket$ = ws$.multiplex(() => subscribeMessage(options, id), () => unSubscribeMessage(options), R.propEq('requestId', id));
        const url = getHttpUrl(options);
        const snapshot$ = socket$.pipe(operators_1.filter(R.propEq('type', types_1.WebsocketRequestType.SUBSCRIBE)), operators_1.tap(() => {
            const base = options.pair.base.symbol;
            const quote = options.pair.quote.symbol;
            debug(`Loading snapshot for market %s-%s.`, base, quote);
        }), operators_1.switchMap(() => axios_1.default.get(url).then(result => result.data)));
        const messages$ = ws$.pipe(operators_1.filter(exports.isOrderEvent));
        // TODO: Make it so, that every time we are fetching a new snapshot (every
        // time we (re-)open the connection), the websocket messages are buffered
        // until the snapshot is emitted.
        const events$ = Rx.merge(snapshot$, messages$);
        return events$.pipe(operators_1.map(R.cond([
            [exports.isSnapshotEvent, data => normalizeSnapshotEvent(options, data)],
            [exports.isNewOrderEvent, data => normalizeNewOrderEvent(options, data.event)],
            [
                exports.isFillOrderEvent,
                data => normalizeFillOrderEvent(options, data.event),
            ],
            [exports.isCancelOrderEvent, data => normalizeCancelOrderEvent(data.event)],
            [exports.isRemoveOrderEvent, data => normalizeRemoveOrderEvent(data.event)],
        ])), operators_1.tap(value => {
            debug(...debug_1.debugEvent(value));
        }));
    }
    catch (error) {
        return Rx.throwError(error);
    }
};
